"""
run few unit test to illustrate implementation of log-normal sv model analytics
"""
import numpy as np
import matplotlib.pyplot as plt
from stochvolmodels import LogSVPricer, LogSvParams, LogsvModelCalibrationType, ConstraintsType, get_btc_test_chain_data

# 1. create instance of pricer
logsv_pricer = LogSVPricer()

# 2. define model params
params = LogSvParams(sigma0=1.0, theta=1.0, kappa1=5.0, kappa2=5.0, beta=0.2, volvol=2.0)

# 3. compute model prices for option slices
model_prices, vols = logsv_pricer.price_slice(params=params,
                                              ttm=0.25,
                                              forward=1.0,
                                              strikes=np.array([0.8, 0.9, 1.0, 1.1]),
                                              optiontypes=np.array(['P', 'P', 'C', 'C']))
print([f"{p:0.4f}, implied vol={v: 0.2%}" for p, v in zip(model_prices, vols)])

# 4. calibrate model to test option chain data
btc_option_chain = get_btc_test_chain_data()
params0 = LogSvParams(sigma0=1.0, theta=1.0, kappa1=2.21, kappa2=2.18, beta=0.15, volvol=2.0)
btc_calibrated_params = logsv_pricer.calibrate_model_params_to_chain(option_chain=btc_option_chain,
                                                                     params0=params0,
                                                                     model_calibration_type=LogsvModelCalibrationType.PARAMS4,
                                                                     constraints_type=ConstraintsType.INVERSE_MARTINGALE)
print(btc_calibrated_params)

# 5. plot model implied vols
logsv_pricer.plot_model_ivols_vs_bid_ask(option_chain=btc_option_chain,
                                         params=btc_calibrated_params)
plt.show()



import numpy as np
from numba import njit
from typing import Union, Tuple


# @njit(cache=False, fastmath=True)
def bracket(ts: np.ndarray, t: float, throw_if_not_found: bool = False) -> int:
    # if 0 in ts:
    #     raise ValueError("0 should be exluded")
    idx0 = -1
    for idx, tk in enumerate(ts):
        if t <= tk:
            idx0 = idx
            break
    if idx0 == -1 and throw_if_not_found:
        raise ValueError('t is not bracketed')
    return idx0


#@njit(cache=False, fastmath=True)
def pw_const(ts: np.ndarray,
             vs: np.ndarray,
             t: float,
             flat_extrapol: bool = False,
             shift: int = 0) -> Union[float, np.ndarray]:
    # if ts[0] < -1e-6 or ts[0] > 1e-6:
    #     raise ValueError('first abscissa must be zero')
    assert shift == 0 or shift == 1
    if ts.shape[0] - shift != vs.shape[0]:
        raise ValueError('abcsissas and ordinates must have same shape')
    value = np.nan
    idx0 = bracket(ts[shift:], t, False)
    value = vs[idx0]
    if flat_extrapol and t >= ts[-1]:
        value = vs[-1]
    return value


# @njit(cache=False, fastmath=True)
def get_default_swap_term_structure(expiry: float, tenor: float):
    freq = 1.0
    return np.arange(expiry, expiry + tenor + freq, freq)  # shift end by freq to include endpoint


@njit(cache=False, fastmath=True)
def get_futures_start_and_pmt(t0: float, lag: float,
                              libor_tenor: float = 0.25) -> Tuple[float, float]:
    start = t0 + lag
    end = start + libor_tenor
    return start, end


# numba version of curve related functions
# @njit(cache=False, fastmath=True)
def df_fast(t: Union[float, np.ndarray], ccy: str = "USD") -> float:
    # flat 0.8% zero rate for JPY and 3.5% for USD
    if ccy == "USD":
        r = 0.043
    elif ccy == "JPY":
        r = 0.008
    elif ccy == "USD_NS":
        lamda = 0.55/12
        beta1 = 0.0436
        beta2 = 0.013
        beta3 = -0.01

        t = np.maximum(t, 1e-4)
        r = beta1 + beta2*(1.0 - np.exp(-lamda * t)) / (lamda * t) + beta3*((1.0 - np.exp(-lamda * t)) / (lamda * t) - np.exp(-lamda * t))


    else:
        raise NotImplementedError

    disc_factor = np.exp(-r * t)
    return disc_factor


def bond_grad(bond_value, B_PX):
    bond_grad = np.zeros((bond_value.size, B_PX.size))
    for j, bj in enumerate(B_PX):
        bond_grad[:, j] = bond_value * bj
    return bond_grad


def swap_grad(numer0: np.ndarray,
              numer1: np.ndarray,
              denumer0: np.ndarray,
              denumer1: np.ndarray) -> np.ndarray:
    # scalar case is exceptional as we don't need to check input consistency
    if numer0.ndim == numer1.ndim == denumer0.ndim == denumer1.ndim:
        swap_grad = numer1 / denumer0 - (numer0 * denumer1) / np.power(denumer0, 2)
        return swap_grad

    assert numer0.ndim == 1 and denumer0.ndim == 1
    assert numer1.ndim == 2 and denumer1.ndim == 2 and np.all(numer1.shape == denumer1.shape)
    swap_grad = np.zeros_like(numer1)
    for j in range(numer1.shape[1]):
        swap_grad[:, j] = numer1[:, j] / denumer0 - (numer0 * denumer1[:, j]) / np.power(denumer0, 2)
    return swap_grad


#############################################################
def generate_ttms_grid(ttms: np.ndarray,
                       nb_pts: int = 11) -> np.ndarray:
    t0 = 0
    t_grid = np.array([0])
    for idx, ttm in enumerate(ttms):
        t_grid0 = np.linspace(t0, ttm, nb_pts)
        t_grid = np.concatenate((t_grid, t_grid0[1:]), axis=None)
        t0 = ttm
    return t_grid


@njit(cache=False, fastmath=True)
def to_yearfrac(d1, d2):
    return d2 - d1
#######################################################


def divide_mc(arr2d, arr1d):
    assert arr2d.ndim == 2 and arr1d.ndim == 1
    res = np.zeros_like(arr2d)
    for j, ann_derj in enumerate(arr2d.T):
        res[:, j] = ann_derj / arr1d
    return res


def prod_mc(arr2d, arr1d):
    assert arr2d.ndim == 2 and arr1d.ndim == 1
    res = np.zeros_like(arr2d)
    for j, ann_derj in enumerate(arr2d.T):
        res[:, j] = ann_derj * arr1d
    return res


# @njit(cache=False, fastmath=True)
def bond(t: float, T: float,
         x: np.ndarray, y: np.ndarray,
         B_PX: np.ndarray, B_PY: np.ndarray,
         ccy: str,
         m: int = 0) -> np.ndarray:
    """return bond value (scalar) or gradient {dB/dx_i, i = 0,..,d} (vector)"""
    assert t <= T
    if x.ndim == 2 or y.ndim == 2:
        # dim of X is # simulations x #factors
        # assume that in matrix X the row is for trials
        # column is for factors
        assert x.shape[0] == y.shape[0]
    assert m == 0 or m == 1
    # because number function cannot return different types, we wrap scalar value into numpy array
    bond_value = np.array([df_fast(T, ccy) / df_fast(t, ccy) * np.exp(-B_PX.dot(np.transpose(x)) - B_PY.dot(np.transpose(y)))])
    # make it just vector, not (1,nsims) matrix
    if x.ndim == 2 or y.ndim == 2:
        bond_value = bond_value[0, :]
    if m == 0:
        return bond_value
    elif m == 1:
        return bond_grad(bond_value, -B_PX)
    else:
        raise NotImplementedError


def swap_rate(ccy: str,
              t: float,
              ts_sw: np.ndarray) -> np.ndarray:
    denumer0 = 0
    for i in range(1, ts_sw.size):
        denumer0 = denumer0 + (ts_sw[i] - ts_sw[i - 1]) * df_fast(ts_sw[i], ccy) / df_fast(t, ccy)

    numer0 = df_fast(ts_sw[0], ccy) / df_fast(t, ccy) - df_fast(ts_sw[-1], ccy) / df_fast(t, ccy)
    value0 = numer0 / denumer0

    return value0


def libor_rate(ccy: str,
               t: float, tenor: float):
    zcb_start = df_fast(t, ccy=ccy)
    zcb_end = df_fast(t+tenor, ccy=ccy)
    libor = 1.0 / tenor * (zcb_start / zcb_end - 1.0)

    return libor


@njit(cache=False, fastmath=True)
def G(k, t, T):
    G_tT = (1.0 - np.exp(-k * (T - t))) / k
    return G_tT


import numpy as np
from numba import njit
from stochvolmodels.pricers.factor_hjm.rate_core import to_yearfrac


@njit(cache=False, fastmath=True)
def init_mean_rev():
    return 0.025


class Discount:
    def __init__(self, currency="USD"):
        self.today = 0
        # flat 0.8% zero rate for JPY and 3.5% for USD
        if currency == "USD":
            self.r = 0.043
        elif currency == "JPY":
            self.r = 0.008
        else:
            raise NotImplementedError

    def df(self, d) -> float:
        year_frac = to_yearfrac(self.today, d)
        disc_factor = np.exp(-self.r * year_frac)
        return disc_factor


def G(t, T):
    k = init_mean_rev()
    G_tT = (1.0 - np.exp(-k * (T - t))) / k
    return G_tT


def bond(t, T, x, y,
         m: int,
         is_mc_mode: bool,
         discount: Discount = None):
    if discount is None:
        discount = Discount()
    if isinstance(t, np.ndarray) and isinstance(x, np.ndarray) and t.shape != x.shape:
        raise ValueError('size of t and x must agree')
    if isinstance(x, np.ndarray) and isinstance(y, np.ndarray) and isinstance(t, np.ndarray) and is_mc_mode:
        raise ValueError('when x and y are vectors, only scalar t is allowed in MC simulations')
    if m < 0 or m > 4:
        raise ValueError('parameter m must be 0,1,2,3,4')
    k = init_mean_rev()
    G = (1.0 - np.exp(-k * (T - t))) / k
    bond_value = discount.df(T) / discount.df(t) * np.exp(-G * x - 0.5 * G ** 2 * y)
    return bond_value * np.power(-G, m)


def annuity(t, ts_sw: np.ndarray, x, y, m,
            discount: Discount = None,
            is_mc_mode: bool = False):
    if discount is None:
        discount = Discount()
    ann = 0
    for i in range(1, ts_sw.size):
        bond_value = bond(t, ts_sw[i], x, y, m, discount=discount, is_mc_mode=is_mc_mode)
        ann = ann + (ts_sw[i] - ts_sw[i-1]) * bond_value
    return ann


def swap_rate(t, ts_sw: np.ndarray, x, y,
              discount: Discount = None,
              is_mc_mode: bool = False):
    if discount is None:
        discount = Discount()
    if (isinstance(x, np.ndarray) and isinstance(y, float)) or (isinstance(x, float) and isinstance(y, np.ndarray)):
        raise ValueError('x and y both must be either scalar or vector')
    if isinstance(x, np.ndarray) and isinstance(y, np.ndarray):
        if x.shape != y.shape:
            raise ValueError('when x and y are vectors, they must have same shape')

    denumer0 = 0
    denumer1 = 0
    denumer2 = 0
    denumer3 = 0
    denumer4 = 0
    for i in range(1, ts_sw.size):
        denumer0 = denumer0 + (ts_sw[i] - ts_sw[i - 1]) * bond(t, ts_sw[i], x, y, 0, discount=discount, is_mc_mode=is_mc_mode)
        denumer1 = denumer1 + (ts_sw[i] - ts_sw[i - 1]) * bond(t, ts_sw[i], x, y, 1, discount=discount, is_mc_mode=is_mc_mode)
        denumer2 = denumer2 + (ts_sw[i] - ts_sw[i - 1]) * bond(t, ts_sw[i], x, y, 2, discount=discount, is_mc_mode=is_mc_mode)
        denumer3 = denumer3 + (ts_sw[i] - ts_sw[i - 1]) * bond(t, ts_sw[i], x, y, 3, discount=discount, is_mc_mode=is_mc_mode)
        denumer4 = denumer4 + (ts_sw[i] - ts_sw[i - 1]) * bond(t, ts_sw[i], x, y, 4, discount=discount, is_mc_mode=is_mc_mode)

    numer0 = bond(t, ts_sw[0], x, y, 0, discount=discount, is_mc_mode=is_mc_mode) - bond(t, ts_sw[-1], x, y, 0, discount=discount, is_mc_mode=is_mc_mode)
    numer1 = bond(t, ts_sw[0], x, y, 1, discount=discount, is_mc_mode=is_mc_mode) - bond(t, ts_sw[-1], x, y, 1, discount=discount, is_mc_mode=is_mc_mode)
    numer2 = bond(t, ts_sw[0], x, y, 2, discount=discount, is_mc_mode=is_mc_mode) - bond(t, ts_sw[-1], x, y, 2, discount=discount, is_mc_mode=is_mc_mode)
    numer3 = bond(t, ts_sw[0], x, y, 3, discount=discount, is_mc_mode=is_mc_mode) - bond(t, ts_sw[-1], x, y, 3, discount=discount, is_mc_mode=is_mc_mode)
    numer4 = bond(t, ts_sw[0], x, y, 4, discount=discount, is_mc_mode=is_mc_mode) - bond(t, ts_sw[-1], x, y, 4, discount=discount, is_mc_mode=is_mc_mode)

    value0 = numer0 / denumer0
    value1 = numer1 / denumer0 - (numer0 * denumer1) / np.power(denumer0, 2)
    value2 = (-2 * numer1 * denumer1) / np.power(denumer0, 2) + numer2 / denumer0 + numer0 * ((2 * np.power(denumer1, 2)) / np.power(denumer0, 3) - denumer2 / np.power(denumer0, 2))
    value3 = (-3 * denumer1 * numer2) / np.power(denumer0, 2) + 3 * numer1 * ((2 * np.power(denumer1, 2)) / np.power(denumer0, 3) - denumer2 / np.power(denumer0, 2)) + numer3 / denumer0 + numer0 * ((-6 * np.power(denumer1, 3)) / np.power(denumer0, 4) + (6 * denumer1 * denumer2) / np.power(denumer0, 3) - denumer3 / np.power(denumer0, 2))
    value4 = (24 * numer0 * np.power(denumer1, 4) - 12 * denumer0 * np.power(denumer1, 2) * (2 * numer1 * denumer1 + 3 * numer0 * denumer2) + 2 * np.power(denumer0, 2) * (6 * np.power(denumer1, 2) * numer2 + 3 * numer0 * np.power(denumer2, 2) + 4 * denumer1 * (3 * numer1 * denumer2 + numer0 * denumer3)) + np.power(denumer0, 4) * numer4 - np.power(denumer0, 3) * (6 * numer2 * denumer2 + 4 * denumer1 * numer3 + 4 * numer1 * denumer3 + numer0 * denumer4)) / np.power(denumer0, 5)

    return value0, value1, value2, value3, value4


def libor_rate(t, t_start: float, t_end: float,
               x, y,
               discount: Discount = None,
               is_mc_mode: bool = False):
    if discount is None:
        discount = Discount()
    if (isinstance(x, np.ndarray) and isinstance(y, float)) or (isinstance(x, float) and isinstance(y, np.ndarray)):
        raise ValueError('x and y both must be either scalar or vector')
    if isinstance(x, np.ndarray) and isinstance(y, np.ndarray):
        if x.shape != y.shape:
            raise ValueError('when x and y are vectors, they must have same shape')

    zcb_start = bond(t, t_start, x, y, 0, discount=discount, is_mc_mode=is_mc_mode)
    zcb_end = bond(t, t_end, x, y, 0, discount=discount, is_mc_mode=is_mc_mode)
    libor = 1.0/(t_end - t_start) * (zcb_start/zcb_end - 1.0)

    return libor




import numpy as np
import matplotlib.pyplot as plt
import stochvolmodels as sv
from stochvolmodels import HestonPricer, HestonParams, OptionChain, BTC_HESTON_PARAMS

pricer = HestonPricer()

# define model params
params = HestonParams(v0=1.0, theta=1.0, kappa=5.0, volvol=1.0, rho=-0.5)

# 1. one price
model_price, vol = pricer.price_vanilla(params=params,
                                        ttm=0.25,
                                        forward=1.0,
                                        strike=1.0,
                                        optiontype='C')
print(f"price={model_price:0.4f}, implied vol={vol: 0.2%}")

# 2. price slice
model_prices, vols = pricer.price_slice(params=params,
                                        ttm=0.25,
                                        forward=1.0,
                                        strikes=np.array([0.9, 1.0, 1.1]),
                                        optiontypes=np.array(['P', 'C', 'C']))
print([f"{p:0.4f}, implied vol={v: 0.2%}" for p, v in zip(model_prices, vols)])

# 3. prices for option chain with uniform strikes
option_chain = OptionChain.get_uniform_chain(ttms=np.array([0.083, 0.25]),
                                             ids=np.array(['1m', '3m']),
                                             strikes=np.linspace(0.9, 1.1, 3))
model_prices, vols = pricer.compute_chain_prices_with_vols(option_chain=option_chain, params=params)
print(model_prices)
print(vols)

# define uniform option chain
option_chain = OptionChain.get_uniform_chain(ttms=np.array([0.083, 0.25]),
                                             ids=np.array(['1m', '3m']),
                                             strikes=np.linspace(0.5, 1.5, 21))
pricer.plot_model_ivols(option_chain=option_chain,
                        params=params)


# define uniform option chain
option_chain = OptionChain.get_uniform_chain(ttms=np.array([0.083, 0.25]),
                                             ids=np.array(['1m', '3m']),
                                             strikes=np.linspace(0.5, 1.5, 21))

# define parameters for bootstrap
params_dict = {'kappa=5': HestonParams(v0=1.0, theta=1.0, kappa=5.0, volvol=1.0, rho=-0.5),
               'kappa=10': HestonParams(v0=1.0, theta=1.0, kappa=10.0, volvol=1.0, rho=-0.5)}
option_slice = option_chain.get_slice(id='1m')
pricer.plot_model_slices_in_params(option_slice=option_slice,
                                   params_dict=params_dict)


btc_option_chain = sv.get_btc_test_chain_data()
btc_calibrated_params = BTC_HESTON_PARAMS
pricer.plot_model_ivols_vs_bid_ask(option_chain=btc_option_chain,
                                   params=btc_calibrated_params)

btc_option_chain = sv.get_btc_test_chain_data()
params0 = HestonParams(v0=0.8, theta=1.0, kappa=5.0, volvol=1.0, rho=-0.5)
btc_calibrated_params = pricer.calibrate_model_params_to_chain(option_chain=btc_option_chain,
                                                               params0=params0,
                                                               constraints_type=sv.ConstraintsType.INVERSE_MARTINGALE)
print(btc_calibrated_params)
pricer.plot_model_ivols_vs_bid_ask(option_chain=btc_option_chain,
                                   params=btc_calibrated_params)
plt.show()



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from numba import njit
from scipy.special import betainc, gamma
from scipy.optimize import fsolve
from typing import Union
from enum import Enum


@njit
def compute_upsilon(vol: float, ttm: float, nu: float) -> float:
    if nu <= 2.0:
        raise ValueError(f"{nu} must be > 2.0")
    return vol*np.sqrt(ttm*(nu-2.0)/nu)


def pdf_tdist(x: Union[np.ndarray, float], mu: float, vol: float, nu: float, ttm: float) -> Union[float, np.ndarray]:
    upsilon = compute_upsilon(vol=vol, ttm=ttm, nu=nu)
    z = (x - mu * ttm) / upsilon
    c = (1.0/np.sqrt(np.pi*nu))*(gamma(0.5*(nu+1.0)) / gamma(0.5*nu))/upsilon
    f = np.power(1.0+np.square(z) / nu, -0.5*(nu+1.0))
    return c*f


def cdf_tdist(x: Union[np.ndarray, float], mu: float, vol: float, nu: float, ttm: float) -> Union[float, np.ndarray]:
    """
    cumulative distribution of cumullative location-scale t-distribution
    cdf = int^{x}_{-\infty} f(u)du
    """
    upsilon = compute_upsilon(vol=vol, ttm=ttm, nu=nu)
    z = (x-mu*ttm) / upsilon
    cdf = 0.5*(1.0 + np.sign(z)*(1.0-betainc(nu/2.0, 0.5, nu/(np.square(z)+nu))))
    return cdf


def cum_mean_tdist(x: Union[np.ndarray, float], mu: float = 0, vol: float = 0.2, nu: float = 3.0, ttm: float = 0.25
                   ) -> Union[float, np.ndarray]:
    """
    cumulative expected value
    h = int^{x}_{-\infty} u f(u)du
    """
    upsilon = compute_upsilon(vol=vol, ttm=ttm, nu=nu)
    z = (x-mu*ttm) / upsilon
    norm = (gamma(0.5*(1.0+nu)) / gamma(0.5*nu))*np.sqrt(nu/np.pi) / (1.0-nu)
    h = mu * cdf_tdist(x, mu=mu, vol=vol, nu=nu, ttm=ttm) + upsilon * norm * np.power(1.0 + np.square(z) / nu, -0.5 * (nu - 1.0))
    return h


def imply_drift_tdist(rf_rate: float = 0.0, vol: float = 0.2, nu: float = 3.0, ttm: float = 0.25) -> float:
    """
    imply drift of t-distribution under risk-neutral measure
    """
    rf_return = (np.exp(rf_rate*ttm) - 1.0)

    def func(mu: float) -> float:
        x_star = -(1.0+ttm*mu)
        return mu * ttm - cdf_tdist(x_star, mu=0.0, vol=vol, nu=nu, ttm=ttm) - cum_mean_tdist(x_star, mu=0.0, vol=vol, nu=nu, ttm=ttm) - rf_return

    mu = fsolve(func, x0=rf_rate, xtol=1e-10)
    return mu[0]


def compute_default_prob_tdist(ttm: float,
                               vol: float,
                               nu: float = 4.5,
                               rf_rate: float = 0.0
                               ) -> Union[float, np.ndarray]:
    """
    imply drift of t-distribution under risk-neutral measure
    """
    risk_neutral_mu = imply_drift_tdist(rf_rate=rf_rate, vol=vol, nu=nu, ttm=ttm)
    x_star = -(1.0+risk_neutral_mu*ttm)
    default_prob = cdf_tdist(x=x_star, mu=0.0, vol=vol, nu=nu, ttm=ttm)
    return default_prob


def compute_forward_tdist(spot: Union[float, np.ndarray],
                          ttm: float,
                          vol: float,
                          nu: float = 4.5,
                          rf_rate: float = 0.0
                          ) -> Union[float, np.ndarray]:
    """
    imply drift of t-distribution under risk-neutral measure
    """
    risk_neutral_mu = imply_drift_tdist(rf_rate=rf_rate, vol=vol, nu=nu, ttm=ttm)
    x_star = -(1.0+risk_neutral_mu*ttm)
    c_1 = cdf_tdist(x=x_star, mu=0.0, vol=vol, nu=nu, ttm=ttm)
    h_1 = cum_mean_tdist(x=x_star, mu=0.0, vol=vol, nu=nu, ttm=ttm)
    forward = spot * ((1.0 + risk_neutral_mu*ttm)*(1.0-c_1)-h_1)
    return forward


def compute_vanilla_price_tdist(spot: Union[float, np.ndarray],
                                strikes: Union[float, np.ndarray],
                                ttm: float,
                                vol: float,
                                nu: float = 4.5,
                                optiontypes: Union[str, np.ndarray] = 'C',
                                rf_rate: float = 0.0,
                                is_compute_risk_neutral_mu: bool = True
                                ) -> Union[float, np.ndarray]:
    """
    option pricer for t-distribution
    """
    discfactor = np.exp(-rf_rate*ttm)
    if is_compute_risk_neutral_mu:
        risk_neutral_mu = imply_drift_tdist(rf_rate=rf_rate, vol=vol, nu=nu, ttm=ttm)
    else:
        risk_neutral_mu = rf_rate
    spot_star = spot*(1.0 + risk_neutral_mu*ttm)
    x_lower_bound = -1.0-risk_neutral_mu*ttm

    def compute(strike_: Union[float, np.ndarray], optiontype_: str) -> float:
        y = strike_ / spot - (1.0 + risk_neutral_mu*ttm)
        c_y = cdf_tdist(x=y, mu=0.0, vol=vol, nu=nu, ttm=ttm)
        h_y = cum_mean_tdist(x=y, mu=0.0, vol=vol, nu=nu, ttm=ttm)
        if optiontype_ == 'C' or optiontype_ == 'IC':
            price_ = (-spot * h_y + (spot_star-strike_)*(1.0-c_y))
        elif optiontype_ == 'P' or optiontype_ == 'IP':
            c_1 = cdf_tdist(x=x_lower_bound, mu=0.0, vol=vol, nu=nu, ttm=ttm)
            h_1 = cum_mean_tdist(x=x_lower_bound, mu=0.0, vol=vol, nu=nu, ttm=ttm)
            price_ = discfactor * ((strike_ - spot_star) * (c_y - c_1) - spot * (h_y - h_1)+strike_*c_1)
        else:
            raise NotImplementedError(f"optiontype")
        return price_

    if isinstance(optiontypes, str):
        price = compute(strikes, optiontypes)
    else:
        price = np.zeros_like(strikes)
        for idx, (strike_, optiontype_) in enumerate(zip(strikes, optiontypes)):
            price[idx] = compute(strike_, optiontype_)
    return price


def infer_implied_vol_tdist(spot: float,
                            ttm: float,
                            strike: float,
                            given_price: float,
                            rf_rate: float = 0.0,
                            optiontype: str = 'C',
                            nu: float = 4.5,
                            tol: float = 1e-12,
                            is_bounds_to_nan: bool = False
                            ) -> float:
    """
    compute normal implied vol
    """
    x1, x2 = 0.05, 10.0  # starting values
    f = compute_vanilla_price_tdist(spot=spot, strikes=strike, ttm=ttm, vol=x1, nu=nu, rf_rate=rf_rate, optiontypes=optiontype) - given_price
    fmid = compute_vanilla_price_tdist(spot=spot, strikes=strike, ttm=ttm, vol=x2, nu=nu, rf_rate=rf_rate, optiontypes=optiontype) - given_price
    if f*fmid < 0.0:
        if f < 0.0:
            rtb = x1
            dx = x2-x1
        else:
            rtb = x2
            dx = x1-x2
        xmid = rtb
        for j in range(0, 100):
            dx = dx*0.5
            xmid = rtb+dx
            fmid = compute_vanilla_price_tdist(spot=spot, strikes=strike, ttm=ttm, vol=xmid, nu=nu, rf_rate=rf_rate, optiontypes=optiontype) - given_price
            if fmid <= 0.0:
                rtb = xmid
            if np.abs(fmid) < tol:
                break
        v1 = xmid
    else:
        if f < 0:
            v1 = x1
        else:
            v1 = x2
    if is_bounds_to_nan:  # in case vol was inferred it will return nan
        if np.abs(v1-x1) < tol or np.abs(v1-x2) < tol:
            v1 = np.nan
    return v1


def infer_tdist_implied_vols_from_model_slice_prices(ttm: float,
                                                     spot: float,
                                                     strikes: np.ndarray,
                                                     optiontypes: np.ndarray,
                                                     model_prices: np.ndarray,
                                                     rf_rate: float,
                                                     nu: float
                                                     ) -> np.ndarray:
    model_vol_ttm = np.zeros_like(strikes)
    for idx, (strike, model_price, optiontype) in enumerate(zip(strikes, model_prices, optiontypes)):
        model_vol_ttm[idx] = infer_implied_vol_tdist(spot=spot, ttm=ttm, rf_rate=rf_rate,
                                                     given_price=model_price,
                                                     strike=strike,
                                                     optiontype=optiontype,
                                                     nu=nu)
    return model_vol_ttm


class UnitTests(Enum):
    PLOT_PDF = 1
    PLOT_CDF = 2
    PLOT_CUM_X = 3
    PLOT_H = 4


def run_unit_test(unit_test: UnitTests):

    import qis as qis

    x = np.linspace(-5.0, 5.0, 20000)
    dx = x[1] - x[0]
    ttm = 1.0
    mu_vols = {'mu=0.0, vol=0.2': (0.0, 0.2),
               'mu=0.2, vol=0.2': (0.2, 0.2),
               'mu=0.2, vol=0.4': (0.2, 0.4)}

    if unit_test == UnitTests.PLOT_PDF:
        pdfs = {}
        for key, mu_vol in mu_vols.items():
            pdf = dx * pdf_tdist(x=x, mu=mu_vol[0], vol=mu_vol[1], nu=3.0, ttm=ttm)
            pdfs[key] = pd.Series(pdf, index=x)
            print(f"{key}: sum={np.sum(pdf)}, mean={np.sum(x*pdf)}, std={np.sqrt(np.sum(np.square(x)*pdf)-np.square(np.sum(x*pdf)))}")
        pdfs = pd.DataFrame.from_dict(pdfs, orient='columns')
        qis.plot_line(df=pdfs)

    elif unit_test == UnitTests.PLOT_CDF:
        pdfs = {}
        cpdfs = {}
        for key, mu_vol in mu_vols.items():
            pdf = dx * pdf_tdist(x=x, mu=mu_vol[0], vol=mu_vol[1], nu=3.0, ttm=ttm)
            cpdf = cdf_tdist(x=x, mu=mu_vol[0], vol=mu_vol[1], nu=3.0, ttm=ttm)
            pdfs[f"{key}_pdf_sum"] = pd.Series(np.cumsum(pdf), index=x)
            cpdfs[f"{key}_cdf"] = pd.Series(cpdf, index=x)
        pdfs = pd.DataFrame.from_dict(pdfs, orient='columns')
        cpdfs = pd.DataFrame.from_dict(cpdfs, orient='columns')
        df = pd.concat([pdfs, cpdfs], axis=1)
        colors = qis.get_n_colors(n=len(mu_vols.keys()))
        qis.plot_line(df=df, colors=2*colors)

    elif unit_test == UnitTests.PLOT_CUM_X:
        pdfs = {}
        cpdfs = {}
        for key, mu_vol in mu_vols.items():
            pdf = dx * pdf_tdist(x=x, mu=mu_vol[0], vol=mu_vol[1], nu=3.0, ttm=ttm)
            cpdf = cum_mean_tdist(x=x, mu=mu_vol[0], vol=mu_vol[1], nu=3.0, ttm=ttm)
            pdfs[f"{key}_h_pdf_sum"] = pd.Series(np.cumsum(x*pdf), index=x)
            cpdfs[f"{key}_t_h"] = pd.Series(cpdf, index=x)
        pdfs = pd.DataFrame.from_dict(pdfs, orient='columns')
        cpdfs = pd.DataFrame.from_dict(cpdfs, orient='columns')
        df = pd.concat([pdfs, cpdfs], axis=1)
        colors = qis.get_n_colors(n=len(mu_vols.keys()))
        qis.plot_line(df=df, colors=2*colors)

    elif unit_test == UnitTests.PLOT_H:
        x = np.linspace(-10.0, 10.0, 2000)
        h = pd.Series(cum_mean_tdist(x=x, mu=0.5, vol=1.0, nu=3.0, ttm=1.0), index=x, name='h')
        qis.plot_line(df=h, xlabel='x')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PLOT_CUM_X

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import numpy as np
from enum import Enum
from typing import Tuple, Optional, Dict, Union, Any
from stochvolmodels.pricers.logsv.affine_expansion import ExpansionOrder, get_expansion_n
from scipy.integrate import solve_ivp
from scipy.integrate._ivp.ivp import OdeResult
from scipy.interpolate import splrep, splev


class UnderlyingType(Enum):
    SWAP = 1
    FUTURES = 2


def compute_logsv_a_mgf_grid(ttm: float,
                             phi_grid: np.ndarray,
                             sigma0: float,
                             q: float,
                             times: np.ndarray,
                             a0: np.ndarray,
                             a1: np.ndarray,
                             kappa0: np.ndarray,
                             kappa1: np.ndarray,
                             kappa2: np.ndarray,
                             beta: np.ndarray,
                             volvol: np.ndarray,
                             b: np.ndarray = None,
                             expansion_order: ExpansionOrder = ExpansionOrder.FIRST,
                             underlying_type: UnderlyingType = UnderlyingType.SWAP,
                             a_t0: Optional[np.ndarray] = None,
                             is_stiff_solver: bool = False,
                             **kwargs
                             ) -> Tuple[np.ndarray, np.ndarray]:
    """
    affine expansion of MGF in rates beta SV model
    """
    if a_t0 is None:
        n_terms = get_expansion_n(expansion_order=expansion_order)
        a_t0 = np.zeros((phi_grid.shape[0], n_terms), dtype=np.complex128)

    if b is None:
        b = np.zeros_like(times)

    a_t1 = solve_a_ode_grid(phi_grid=phi_grid, ttm=ttm, q=q, times=times, a0=a0, a1=a1, kappa0=kappa0,
                            kappa1=kappa1, kappa2=kappa2, beta=beta, volvol=volvol, b=b,
                            a_t0=a_t0, expansion_order=expansion_order, is_stiff_solver=is_stiff_solver,
                            underlying_type=underlying_type)
    y = sigma0 - q
    if expansion_order == ExpansionOrder.FIRST:
        ys = np.array([1.0, y, y * y])  # params.v0 - params.theta
    elif expansion_order == ExpansionOrder.SECOND:
        y2 = y * y
        ys = np.array([1.0, y, y2, y2 * y, y2 * y2])
    else:
        raise NotImplementedError
    ys = ys + 1j * 0.0  # simple way to convert to complex array
    log_mgf = a_t1 @ ys
    return a_t1, log_mgf


# cannot use @njit(cache=False, fastmath=True) when using solve_ode_for_a with solve_ivp
def solve_a_ode_grid(phi_grid: np.ndarray,
                     ttm: float,
                     q: float,
                     times: np.ndarray,
                     a0: np.ndarray,
                     a1: np.ndarray,
                     kappa0: np.ndarray,
                     kappa1: np.ndarray,
                     kappa2: np.ndarray,
                     beta: np.ndarray,
                     volvol: np.ndarray,
                     b: np.ndarray = None,
                     a_t0: Optional[np.ndarray] = None,
                     is_stiff_solver: bool = False,
                     expansion_order: ExpansionOrder = ExpansionOrder.FIRST,
                     underlying_type: UnderlyingType = UnderlyingType.SWAP,
                     ) -> np.ndarray:
    """
    solve ode for range phi
    """

    def fun(phi: np.complex128, a_t0: np.ndarray):
        oderesult = solve_ode_for_a(ttm=ttm, q=q, times=times, a0=a0, a1=a1, kappa0=kappa0, kappa1=kappa1,
                                    kappa2=kappa2, beta=beta, volvol=volvol, b=b,
                                    phi=phi, expansion_order=expansion_order,
                                    is_stiff_solver=is_stiff_solver, underlying_type=underlying_type,
                                    a_t0=a_t0)
        return oderesult

    # pre-allocate array for solutions
    a_t1 = np.zeros((phi_grid.shape[0], get_expansion_n(expansion_order)), dtype=np.complex128)
    for idx, phi in enumerate(phi_grid):
        a_t1[idx, :] = fun(phi, a_t0[idx, :]).y[:, -1]
        # print(f"Time for MGF calculation at {phi}: {toc - tic:0.4f} seconds")
    return a_t1


def solve_ode_for_a(ttm: float,
                    q: float,
                    times: np.ndarray,
                    a0: np.ndarray,
                    a1: np.ndarray,
                    kappa0: np.ndarray,
                    kappa1: np.ndarray,
                    kappa2: np.ndarray,
                    beta: np.ndarray,
                    volvol: np.ndarray,
                    phi: np.complex128,
                    b: np.ndarray = None,
                    a_t0: Optional[np.ndarray] = None,
                    expansion_order: ExpansionOrder = ExpansionOrder.FIRST,
                    is_stiff_solver: bool = False,
                    underlying_type: UnderlyingType = UnderlyingType.SWAP,
                    dense_output: bool = False
                    ) -> OdeResult:
    """
    solve ode for given scalar phi
    """
    a0 = [splrep(times, a0[:,j], s=0) for j in range(a0.shape[1])]
    a1 = splrep(times, a1)
    kappa0 = splrep(times, kappa0)
    kappa1 = splrep(times, kappa1)
    kappa2 = splrep(times, kappa2)
    beta = [splrep(times, beta[:,j]) for j in range(beta.shape[1])]
    volvol = splrep(times, volvol)
    b = splrep(times, b)
    if is_stiff_solver:
        ode_sol = solve_ivp(fun=func_rhs, t_span=(0.0, ttm), y0=a_t0,
                            args=(ttm, q, times, a0, a1, kappa0, kappa1, kappa2, beta, volvol, b, phi, underlying_type, expansion_order),
                            method='BDF',
                            jac=func_rhs_jac,
                            dense_output=dense_output)
    else:
        ode_sol = solve_ivp(fun=func_rhs, t_span=(0.0, ttm), y0=a_t0,
                            args=(ttm, q, times, a0, a1, kappa0, kappa1, kappa2, beta, volvol, b, phi, underlying_type, expansion_order),
                            dense_output=dense_output)
    return ode_sol


# AFFINE EXPANSION FOR SWAPTIONS and FUTURES OPTIONS
# @njit(cache=False, fastmath=True) # TODO: uncomment
def func_rhs(tau: float,  # for ode solver compatibility
             A0: np.ndarray,
             ttm: float,
             q: float,
             times: np.ndarray,
             a0: Any,
             a1: Any,
             kappa0: Any,
             kappa1: Any,
             kappa2: Any,
             beta: Any,
             volvol: Any,
             b: np.ndarray,
             phi: np.complex128,
             underlying_type: UnderlyingType,
             expansion_order: ExpansionOrder) -> np.ndarray:
    """
    returns rhs of the ODEs
    """
    n = A0.shape[0]
    quadratic = np.zeros(n, dtype=np.complex128)
    # a_i = pw_const(ts=times, vs=a, t=ttm - tau)
    # delta_i = pw_const(ts=times, vs=delta, t=ttm - tau)
    # kappa0_i = pw_const(ts=times, vs=kappa0, t=ttm - tau)
    # kappa1_i = pw_const(ts=times, vs=kappa1, t=ttm - tau)
    # kappa2_i = pw_const(ts=times, vs=kappa2, t=ttm - tau)
    # beta_i = pw_const(ts=times, vs=beta, t=ttm - tau)
    # volvol_i = pw_const(ts=times, vs=volvol, t=ttm - tau)
    a0_i = np.array([float(splev(ttm-tau, a0[j])) for j in range(len(a0))])
    a1_i = float(splev(ttm-tau, a1))
    b_i = float(splev(ttm-tau, b))
    kappa0_i = float(splev(ttm-tau, kappa0))
    kappa1_i = float(splev(ttm-tau, kappa1))
    kappa2_i = float(splev(ttm-tau, kappa2))
    beta_i = np.array([float(splev(ttm-tau, beta[j])) for j in range(len(beta))])
    volvol_i = float(splev(ttm-tau, volvol))

    M, L, H = func_a_ode_quadratic_terms(q=q, a0=a0_i, a1=a1_i, kappa0=kappa0_i, kappa1=kappa1_i, kappa2=kappa2_i,
                                         beta=beta_i, volvol=volvol_i, phi=phi, b=b_i,
                                         underlying_type=underlying_type, expansion_order=expansion_order)

    for n_ in np.arange(n):
        quadratic[n_] = A0.T @ M[n_] @ A0
    rhs = quadratic + L @ A0 + H
    return rhs


# @njit(cache=False, fastmath=True)  # TODO: uncomment
def func_rhs_jac(tau: float,  # for ode solver compatibility
                 A0: np.ndarray,
                 ttm: float,
                 q: float,
                 times: np.ndarray,
                 a0: np.ndarray,
                 a1: np.ndarray,
                 kappa0: np.ndarray,
                 kappa1: np.ndarray,
                 kappa2: np.ndarray,
                 beta: np.ndarray,
                 volvol: np.ndarray,
                 b: np.ndarray,
                 phi: np.complex128,
                 underlying_type: UnderlyingType,
                 expansion_order: ExpansionOrder) -> np.ndarray:
    """
    returns rhs jacobian evaluation using matrices for the quadratic form A_t = A.T@M@A + L@A + H
    """
    n = A0.shape[0]
    quadratic = np.zeros((n, n), dtype=np.complex128)
    # a_i = pw_const(ts=times, vs=a, t=ttm - tau)
    # delta_i = pw_const(ts=times, vs=delta, t=ttm - tau)
    # kappa0_i = pw_const(ts=times, vs=kappa0, t=ttm - tau)
    # kappa1_i = pw_const(ts=times, vs=kappa1, t=ttm - tau)
    # kappa2_i = pw_const(ts=times, vs=kappa2, t=ttm - tau)
    # beta_i = pw_const(ts=times, vs=beta, t=ttm - tau)
    # volvol_i = pw_const(ts=times, vs=volvol, t=ttm - tau)
    a0_i = np.array([float(splev(ttm - tau, a0[j])) for j in range(len(a0))])
    a1_i = float(splev(ttm - tau, a1))
    b_i = float(splev(ttm - tau, b))
    kappa0_i = float(splev(ttm - tau, kappa0))
    kappa1_i = float(splev(ttm - tau, kappa1))
    kappa2_i = float(splev(ttm - tau, kappa2))
    beta_i = np.array([float(splev(ttm - tau, beta[j])) for j in range(len(beta))])
    volvol_i = float(splev(ttm - tau, volvol))

    M, L, H = func_a_ode_quadratic_terms(q=q, a0=a0_i, a1=a1_i, kappa0=kappa0_i, kappa1=kappa1_i, kappa2=kappa2_i,
                                         beta=beta_i, volvol=volvol_i, phi=phi, b=b_i,
                                         underlying_type=underlying_type, expansion_order=expansion_order)
    for n_ in np.arange(n):
        quadratic[n_, :] = 2.0 * M[n_] @ A0
    rhs = quadratic + A0
    return rhs

# @njit(cache=False, fastmath=True) # TODO: uncomment
def func_a_ode_quadratic_terms(q: float,
                               a0: Union[float, np.ndarray],
                               a1: float,
                               kappa0: float,
                               kappa1: float,
                               kappa2: float,
                               beta: Union[float, np.ndarray],
                               volvol: float,
                               b: float,
                               phi: np.complex128,
                               underlying_type: UnderlyingType,
                               expansion_order: ExpansionOrder
                               ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Matrices for the quadratic form A_t = A.T@M@A + L@A + H
    """
    q2 = q * q
    # if isinstance(beta, float):
    #     vartheta2 = beta * beta + volvol * volvol
    # elif isinstance(beta, np.ndarray):
    #     vartheta2 = np.dot(beta, beta) + volvol * volvol
    # else:
    #     raise NotImplementedError(f"beta can be either scalar or vector")
    vartheta2 = np.dot(beta, beta) + volvol * volvol
    qv = q * vartheta2
    qv2 = q2 * vartheta2

    # if underlying_type == UnderlyingType.FUTURES and expansion_order == ExpansionOrder.SECOND:
    #     raise NotImplementedError(f"Second order expansion for ED options is not supported yet")

    # if isinstance(a0, float):
    #     a_prod_beta = a0 * beta
    #     a_prod_a = a0 * a0
    if underlying_type == UnderlyingType.FUTURES:
        a_prod_beta = np.dot(a0, beta) + a1 * volvol
        # a_prod_beta = np.dot(a, beta)
        a_prod_a = np.dot(a0, a0) + a1*a1
    elif underlying_type == UnderlyingType.SWAP:
        a_prod_beta = np.dot(a0, beta)
        a_prod_a = np.dot(a0, a0)
    else:
        raise NotImplementedError(f"beta can be either scalar or vector")

    # fill Ms: M should be of same type as L and H for numba, even though they are real
    # utilize that M is symmetric
    # although tuple is more intuitive for M, because of numba we need to make it a tensor
    n = get_expansion_n(expansion_order=expansion_order)
    M = np.zeros((n, n, n), dtype=np.complex128)
    M[0, 1, 1] = 0.5 * qv2

    M[1, 1, 1] = qv
    M[1, 1, 2] = M[1, 2, 1] = qv2

    M[2, 1, 1] = 0.5 * vartheta2
    M[2, 2, 2] = 2.0 * qv2
    M[2, 2, 1] = M[2, 1, 2] = 2.0 * qv

    # if expansion_order == ExpansionOrder.SECOND:
    #     raise ValueError("Second-order expansion not supported. TODO")
    if expansion_order == ExpansionOrder.SECOND:
        M[2, 1, 3] = M[2, 3, 1] = 1.5*qv2

        M[3, 2, 2] = 4.0*qv
        M[3, 1, 2] = M[3, 2, 1] = vartheta2
        M[3, 1, 3] = M[3, 3, 1] = 3.0*qv
        M[3, 1, 4] = M[3, 4, 1] = 2.0 * qv2
        M[3, 2, 3] = M[3, 3, 2] = 3.0 * qv2

        M[4, 2, 2] = 2.0 * vartheta2
        M[4, 3, 3] = 4.5 * qv2
        M[4, 1, 3] = M[4, 3, 1] = 1.5 * vartheta2
        M[4, 1, 4] = M[4, 4, 1] = 4.0 * qv
        M[4, 2, 3] = M[4, 3, 2] = 6.0 * qv
        M[4, 2, 4] = M[4, 4, 2] = 4.0 * qv2

    # fills Ls
    L = np.zeros((n, n), dtype=np.complex128)
    L[0, 1] = kappa0 - q2 * a_prod_beta * phi
    L[0, 2] = qv2
    L[1, 1] = -kappa1 - 2.0 * q * a_prod_beta * phi
    L[1, 2] = 2.0 * (kappa0 + qv - q2 * a_prod_beta * phi)
    L[2, 1] = -kappa2 - a_prod_beta * phi
    L[2, 2] = vartheta2 - 2.0 * kappa1 - 4.0 * q * a_prod_beta * phi

    # if expansion_order == ExpansionOrder.SECOND:
    #     raise ValueError("Second-order expansion not supported. TODO")
    if expansion_order == ExpansionOrder.SECOND:
        L[1, 3] = 3.0*qv2
        L[2, 3] = 3.0 * (kappa0 - q2 * a_prod_beta * phi + 2.0 * qv)
        L[2, 4] = 6.0 * qv2
        L[3, 2] = -2.0 * (kappa2 + a_prod_beta * phi)
        L[3, 3] = 3.0 * (vartheta2 - kappa1 - 2.0 * q * a_prod_beta * phi)
        L[3, 4] = 4.0 * (3.0 * qv + kappa0 - q2 * a_prod_beta * phi)
        L[4, 3] = -3.0 * (kappa2 + a_prod_beta * phi)
        L[4, 4] = 2.0 * (3.0 * vartheta2 - 2.0 * kappa1 - 4.0 * q * a_prod_beta * phi)

    # fill Hs
    H = np.zeros(n, dtype=np.complex128)

    H[0] = 0.5 * q2 * phi * (2.0 * b + a_prod_a * phi)
    H[1] = q * phi * (2.0 * b + a_prod_a * phi)
    H[2] = 0.5 * phi * (2.0 * b + a_prod_a * phi)

    return M, L, H



import numpy as np
import matplotlib.pyplot as plt
from stochvolmodels import HestonPricer, HestonParams, OptionChain

# define parameters for bootstrap
params_dict = {'rho=0.0': HestonParams(v0=0.2**2, theta=0.2**2, kappa=4.0, volvol=0.75, rho=0.0),
               'rho=-0.4': HestonParams(v0=0.2**2, theta=0.2**2, kappa=4.0, volvol=0.75, rho=-0.4),
               'rho=-0.8': HestonParams(v0=0.2**2, theta=0.2**2, kappa=4.0, volvol=0.75, rho=-0.8)}

# get uniform slice
option_chain = OptionChain.get_uniform_chain(ttms=np.array([0.25]), ids=np.array(['3m']), strikes=np.linspace(0.8, 1.15, 20))
option_slice = option_chain.get_slice(id='3m')

# run pricer
pricer = HestonPricer()
pricer.plot_model_slices_in_params(option_slice=option_slice, params_dict=params_dict)

plt.show()



"""
analytics for vol and QV moments computation
"""
# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from numpy import linalg as la
from scipy import linalg as sla
from enum import Enum
# project
from stochvolmodels.pricers.logsv.logsv_params import LogSvParams
from stochvolmodels.utils.funcs import set_seed


VOLVOL = 1.75

DRIFT_PARAMS = {'$(\kappa_{1}=4, \kappa_{2}=0)$': LogSvParams(sigma0=1.0, theta=1.0, kappa1=4.0, kappa2=0.0, beta=0.0, volvol=VOLVOL),
                '$(\kappa_{1}=4, \kappa_{2}=4)$': LogSvParams(sigma0=1.0, theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=VOLVOL),
                '$(\kappa_{1}=4, \kappa_{2}=8)$': LogSvParams(sigma0=1.0, theta=1.0, kappa1=4.0, kappa2=8.0, beta=0.0, volvol=VOLVOL)}


def compute_analytic_vol_moments(params: LogSvParams,
                                 t: float = 1.0,
                                 n_terms: int = 4,
                                 is_qvar: bool = False
                                 ) -> np.ndarray:

    lambda_m = params.get_vol_moments_lambda(n_terms=n_terms)

    y = params.sigma0 - params.theta
    y0 = np.zeros(n_terms)
    for n in range(0, n_terms):
        y0[n] = np.power(y, n+1)

    if np.isclose(np.abs(t), 0.0):
        return y0

    rhs = np.zeros(n_terms)
    rhs[1] = params.vartheta2*params.theta2

    if is_qvar:  # need flat boundary condition
        rhs[-1] = -n_terms*params.kappa2*np.power(y, n_terms+1)
    else:
        rhs[-1] = -n_terms*params.kappa2*np.power(y, n_terms+1)

    i_m = la.inv(lambda_m)
    is_expm = True
    if is_expm:
        e_m = sla.expm(lambda_m*t)
        m_rhs = i_m @ (e_m - np.eye(n_terms))
    else:
        w, v = la.eig(lambda_m)
        v_inv = la.inv(v)
        e_m = np.real(v @ np.diag(np.exp(w*t)) @ v_inv)
        m_rhs = np.real(v @ np.diag(np.reciprocal(w)*(np.exp(w*t)-np.ones(n_terms))) @ v_inv)
        # m_rhs = i_m @ (e_m - np.eye(n_terms))

    if is_qvar:
        sol1 = m_rhs @ y0
        intm2 = i_m @ (m_rhs-t*np.eye(n_terms))
        sol2 = intm2 @ rhs
    else:
        sol1 = e_m @ y0
        sol2 = m_rhs @ rhs

    sol = sol1 + sol2
    return sol


def compute_analytic_qvar(params: LogSvParams,
                          ttm: float = 1.0,
                          n_terms: int = 4
                          ) -> float:
    """
    compute expected value [ (1/T) int^T_0 sigma^2_t dt]
    """
    if np.isclose(ttm, 0.0):
        qvar = np.square(params.sigma0)
    else:
        int_moments = compute_analytic_vol_moments(params=params, t=ttm, n_terms=n_terms, is_qvar=True)
        qvar = (int_moments[1] + 2.0*params.theta*int_moments[0]) / ttm + params.theta2
    return qvar


def compute_vol_moments_t(params: LogSvParams,
                          ttm: np.ndarray,
                          n_terms: int = 4,
                          is_print: bool = False
                          ) -> np.ndarray:
    moments = np.zeros((len(ttm), n_terms))
    for idx, t_ in enumerate(ttm):
        moments_ = compute_analytic_vol_moments(t=t_, params=params, n_terms=n_terms)
        if is_print:
            print(f"t={t_}: {moments_}")
        moments[idx, :] = moments_
    return moments


def compute_expected_vol_t(params: LogSvParams,
                           t: np.ndarray,
                           n_terms: int = 4,
                           ) -> np.ndarray:
    ev = np.zeros(len(t))
    for idx, t_ in enumerate(t):
        moments = compute_analytic_vol_moments(t=t_, params=params, n_terms=n_terms)
        ev[idx] = moments[0] + params.theta
    return ev


def compute_sqrt_qvar_t(params: LogSvParams, t: np.ndarray, n_terms: int = 4) -> np.ndarray:
    ev = np.zeros(len(t))
    for idx, t_ in enumerate(t):
        ev[idx] = np.sqrt(compute_analytic_qvar(ttm=t_, params=params, n_terms=n_terms))
    return ev


def fit_model_vol_backbone_to_varswaps(log_sv_params: LogSvParams,
                                       varswap_strikes: pd.Series,
                                       n_terms: int = 4,
                                       verbose: bool = False
                                       ) -> pd.Series:
    """
    fit model eta so that model reproduces quadratic var
    """
    ttms = varswap_strikes.index.to_numpy()
    market_qvar_dt = ttms * np.square(varswap_strikes.to_numpy())
    # compute model qvars
    model_forwards = np.array([compute_analytic_qvar(params=log_sv_params, ttm=ttm, n_terms=n_terms) for ttm in ttms])
    model_qvar_dt = model_forwards*ttms
    model_eta = np.ones_like(ttms)
    for idx, ttm in enumerate(ttms):
        if idx == 0:
            model_eta[idx] = market_qvar_dt[idx] / model_qvar_dt[idx]
        else:
            model_eta[idx] = (market_qvar_dt[idx]-market_qvar_dt[idx-1]) / (model_qvar_dt[idx]-model_qvar_dt[idx-1])
    # model_eta = np.where(model_eta > 0.0, np.sqrt(model_eta), 1.0)
    model_eta = np.where(model_eta > 0.0, model_eta, 1.0)
    # adhoc adjustemnt for now
    model_eta = np.where(ttms < 0.06, np.sqrt(model_eta), model_eta)

    model_eta = pd.Series(model_eta, index=ttms)
    if verbose:
        varswap_strikes = np.sqrt(varswap_strikes.to_frame('vars_swap strikes'))
        varswap_strikes['market_qvar_dt'] = market_qvar_dt
        varswap_strikes['model_qvar_dt'] = model_qvar_dt
        varswap_strikes['model_eta'] = model_eta
        print(f"vars_swaps\n{varswap_strikes}")
    return model_eta


class UnitTests(Enum):
    VOL_MOMENTS = 1
    EXPECTED_VOL = 2
    EXPECTED_QVAR = 3
    VOL_BACKBONE = 4


def run_unit_test(unit_test: UnitTests):

    from stochvolmodels.pricers.logsv_pricer import LogSVPricer
    logsv_pricer = LogSVPricer()

    n_terms = 4
    nb_path = 200000
    ttm = 1.0
    params = LogSvParams(sigma0=1.0, theta=1.0, kappa1=4.0, kappa2=4.0, beta=0.0, volvol=1.75)
    params.assert_vol_moments_stability(n_terms=n_terms)
    set_seed(8)  # 8
    sigma_t, grid_t = logsv_pricer.simulate_vol_paths(ttm=ttm, params=params, nb_path=nb_path)

    if unit_test == UnitTests.VOL_MOMENTS:

        mcs = []
        for n in np.arange(n_terms):
            if n > 0:
                m_n = np.power(sigma_t-params.theta, n+1)
            else:
                m_n = sigma_t - params.theta
            mc_mean, mc_std = np.mean(m_n, axis=1), np.std(sigma_t, axis=1) / np.sqrt(nb_path)
            mc = pd.Series(mc_mean, index=grid_t, name=f"MC m{n+1}")
            mc_m = pd.Series(mc_mean-1.96*mc_std, index=grid_t, name='MC-cd')
            mc_p = pd.Series(mc_mean+1.96*mc_std, index=grid_t, name='MC+cd')
            mcs.append(mc)
        analytic_vol_moments = compute_vol_moments_t(params=params, ttm=grid_t, n_terms=n_terms)
        analytic_vol_moments = pd.DataFrame(analytic_vol_moments, index=grid_t, columns=[f"m{n+1}" for n in range(n_terms)])
        mcs = pd.concat(mcs, axis=1)

        df = pd.concat([analytic_vol_moments, mcs], axis=1)
        print(df)
        df.plot()

    elif unit_test == UnitTests.EXPECTED_VOL:

        mc_mean, mc_std = np.mean(sigma_t, axis=1), np.std(sigma_t, axis=1) / np.sqrt(nb_path)
        mc = pd.Series(mc_mean, index=grid_t, name='MC')
        mc_m = pd.Series(mc_mean-1.96*mc_std, index=grid_t, name='MC-cd')
        mc_p = pd.Series(mc_mean+1.96*mc_std, index=grid_t, name='MC+cd')

        analytic_vol_moments = compute_expected_vol_t(params=params, t=grid_t, n_terms=n_terms)
        analytic_vol_moments = pd.Series(analytic_vol_moments, index=grid_t, name='Analytic')

        df = pd.concat([analytic_vol_moments, mc, mc_m, mc_p], axis=1)
        print(df)
        df.plot()

    elif unit_test == UnitTests.EXPECTED_QVAR:

        q_var = pd.DataFrame(np.square(sigma_t)).expanding(axis=0).mean().to_numpy()
        mc_mean = np.sqrt(np.mean(q_var, axis=1))
        mc_std = np.std(q_var, axis=1) / np.sqrt(nb_path)
        mc = pd.Series(mc_mean, index=grid_t, name='MC')
        mc_m = pd.Series(mc_mean-1.96*mc_std, index=grid_t, name='MC-cd')
        mc_p = pd.Series(mc_mean+1.96*mc_std, index=grid_t, name='MC+cd')

        analytic_vol_moments = compute_sqrt_qvar_t(params=params, t=grid_t, n_terms=n_terms)
        analytic_vol_moments = pd.Series(analytic_vol_moments, index=grid_t, name='Analytic')

        df = pd.concat([analytic_vol_moments, mc, mc_m, mc_p], axis=1)
        with sns.axes_style('darkgrid'):
            fig, ax = plt.subplots(1, 1, figsize=(18, 10), tight_layout=True)
            sns.lineplot(data=analytic_vol_moments, dashes=False, ax=ax)
            ax.errorbar(x=df.index[::5], y=mc_mean[::5], yerr=mc_std[::5], fmt='o', color='green', capsize=8)

    elif unit_test == UnitTests.VOL_BACKBONE:
        fit_model_vol_backbone_to_varswaps(log_sv_params=params,
                                           varswap_strikes=pd.Series([1.0, 1.0], index=[1.0 / 12., 2 / 12.0]),
                                           verbose=True)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.VOL_BACKBONE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import numpy as np
from typing import Union, Tuple
from dataclasses import dataclass
from numba.typed import List
from scipy.integrate import solve_ivp

from stochvolmodels.pricers.model_pricer import ModelParams
from stochvolmodels.pricers.factor_hjm.rate_factor_basis import Cheyette1D, NelsonSiegel, CheyettePEND
from stochvolmodels.pricers.factor_hjm.rate_core import pw_const, G, bracket, get_default_swap_term_structure, \
    generate_ttms_grid


@dataclass
class TermStructure:
    ts: np.ndarray
    xs: np.ndarray
    flat_extrapol: bool = False

    def __post_init__(self):
        """
        check that market data dimensions are consistent
        """
        if self.ts.ndim != 1:
            raise ValueError('ts must have 1 dimension')
        if self.xs.ndim != 1 and self.xs.ndim != 2:
            raise ValueError('xs must have dimension of one or two')
        if self.ts.shape[0] - 1 != self.xs.shape[0]:
            raise ValueError('abcsissas and ordinates must have same shape')


    # @njit(cache=False, fastmath=True) # TODO: cannot make it numba as it is member function
    def pw_const(self, t: float) -> float:
        """
        Piecewise constant interpolation with possibly flat extrapolation
        """
        return pw_const(self.ts, self.xs, t, self.flat_extrapol, shift=1)

    # @njit(cache=False, fastmath=True) # TODO: cannot make it numba as it is member function
    def interpolate(self, times: np.ndarray) -> np.ndarray:
        interp_values = np.ones_like(times) * np.nan
        for idx, t in enumerate(times):
            interp_values[idx] = self.pw_const(t)
        return interp_values

    @classmethod
    def create_from_scalar(cls, ts: np.ndarray, xs: float, flat_extrapol: bool = False):
        return TermStructure(ts=ts, xs=np.ones_like(ts[1:]) * xs, flat_extrapol=flat_extrapol)

    @classmethod
    def create_multi_fact_from_vec(cls, ts: np.ndarray, xs: np.ndarray, flat_extrapol: bool = False):
        assert xs.ndim == 1
        # when value is 1D vector, its value are for each factor
        # we have as many columns as factors
        # as many rows as time points
        xs_ = np.zeros((ts[1:].size, xs.size))
        for j, val in enumerate(xs):
            xs_[:, j] = val * np.ones_like(ts[1:])
        return TermStructure(ts=ts, xs=xs_, flat_extrapol=flat_extrapol)



@dataclass
class RateLogSvParamsBase(ModelParams):
    """
    Implementation of model params class
    """
    sigma0: float
    theta: float
    kappa1: float
    kappa2: float
    alpha: TermStructure  # term-structure
    b: TermStructure  # term-structure
    beta: TermStructure  # term-structure
    volvol: TermStructure  # term-structure
    ccy: str    # currency in discounting
    basis: Cheyette1D  # basis for 1D Cheyette
    term: float  # key swap tenor
    q: float = None


@dataclass
class RateLogSvParams(RateLogSvParamsBase):
    def calc_mean_states(self,
                         expiry: float,
                         t_grid: np.ndarray) -> (np.ndarray, np.ndarray):
        mrv_r = self.basis.meanrev
        ts_sw = get_default_swap_term_structure(expiry=expiry, tenor=self.term)

        def func_mean_rhs(t: float,
                          arg: np.ndarray,
                          kappa_r: float,
                          ts_sw_: np.ndarray,
                          a: TermStructure,
                          beta: TermStructure,
                          kappa1: float,
                          kappa2: float,
                          theta: float) -> np.ndarray:
            res = np.zeros((3,))
            x, y, sigma = arg[0], arg[1], arg[2]
            a_t = self.alpha.pw_const(t)
            beta_t = self.beta.pw_const(t)
            loga_der = 1.0 / self.basis.annuity(t, ts_sw_, x, y, ccy=self.ccy, m=0) * self.basis.annuity(t, ts_sw_, x, y, ccy=self.ccy, m=1)
            res[0] = y - kappa_r * x + loga_der * a_t ** 2 * sigma ** 2
            res[1] = a_t ** 2 * sigma ** 2 - 2.0 * kappa_r * y
            res[2] = (kappa1 + kappa2 * sigma) * (theta - sigma) + a_t * beta_t * loga_der * sigma ** 2

            return res

        A_t0 = np.array([0.0, 0.0, self.sigma0])  # TODO: review initial conditions for x,y
        # solve ODE with initial conditions x(0), y(0)
        oderesult = solve_ivp(fun=func_mean_rhs, t_span=(0, expiry), t_eval=t_grid, y0=A_t0,
                              args=(mrv_r, ts_sw, self.alpha, self.beta, self.kappa1, self.kappa2, self.theta),
                              dense_output=False)
        (mx, my) = oderesult.y[:2, :]
        return mx, my

    def transform_QA_params(self,
                            expiry: float,
                            tenor: float,
                            t_grid: np.ndarray):

        theta = self.theta
        kappa1 = self.kappa1
        kappa2 = self.kappa2
        if self.q is None:
            self.q = self.theta
        q = self.q

        # in fact, tenor in argument is just dummy variable
        # but still need to check for consistency
        assert tenor == self.term
        ts_sw = get_default_swap_term_structure(expiry=expiry, tenor=tenor)
        if expiry not in t_grid:
            raise ValueError(f"expiry must be in grid")
        idx_ttm = np.where(t_grid == expiry)[0][0]
        t_grid = t_grid[:idx_ttm + 1]

        mx_grid, my_grid = self.calc_mean_states(expiry, t_grid)

        swap_der1 = np.ones_like(t_grid)
        ann = np.ones_like(t_grid)
        ann_der1 = np.ones_like(t_grid)
        for idx, (t, mx, my) in enumerate(zip(t_grid, mx_grid, my_grid)):
            swap_der1[idx] = self.basis.swap_rate(t, ts_sw, mx, my, ccy=self.ccy)[1]
            ann[idx] = self.basis.annuity(t, ts_sw, mx, my, ccy=self.ccy, m=0)
            ann_der1[idx] = self.basis.annuity(t, ts_sw, mx, my, ccy=self.ccy, m=1)

        loga_der = ann_der1 / ann

        # interpolate a, beta, vol-vol from coarse grid to finer one
        alpha_interp = self.alpha.interpolate(t_grid)
        beta_interp = self.beta.interpolate(t_grid)
        volvol_interp = self.volvol.interpolate(t_grid)

        a = alpha_interp * swap_der1  # coordinate-wise product of arrays
        beta2 = beta_interp * loga_der  # coordinate-wise product of arrays

        term0 = alpha_interp * beta2 * (q ** 2) + (theta - q) * kappa1 + (theta - q) * kappa2 * q
        term1 = kappa1 - kappa2 * q + 2.0 * (kappa2 - alpha_interp * beta2) * q - (theta - q) * kappa2
        term2 = kappa2 - alpha_interp * beta2

        return a, term0, term1, term2, beta_interp, volvol_interp, ts_sw

    def reduce(self, idx: int):
        param = RateLogSvParams(sigma0=self.sigma0,
                                theta=self.theta,
                                kappa1=self.kappa1,
                                kappa2=self.kappa2,
                                alpha=TermStructure(self.alpha.ts[:idx + 1], self.alpha.xs[:idx]),
                                b=TermStructure(self.b.ts[:idx + 1], self.b.xs[:idx]),
                                beta=TermStructure(self.beta.ts[:idx + 1], self.beta.xs[:idx]),
                                volvol=TermStructure(self.volvol.ts[:idx + 1], self.volvol.xs[:idx]),
                                ccy=self.ccy,
                                basis=self.basis,
                                term=self.term)
        return param


    def transform_QT_params(self,
                            expiry: float,
                            t_start: float,
                            t_end: float,
                            t_grid: np.ndarray):
        theta = self.theta
        kappa1 = self.kappa1
        kappa2 = self.kappa2
        self.q = self.theta
        q = self.q

        # interpolate a, beta, vol-vol from coarse grid to finer one
        alpha_interp = self.alpha.interpolate(t_grid)
        beta_interp = self.beta.interpolate(t_grid)
        volvol_interp = self.volvol.interpolate(t_grid)

        k = self.basis.meanrev
        G_t_T = G(k, t_grid, expiry)
        G_start_end = G(k, t_start, t_end)


        a = alpha_interp * G_start_end * np.exp(-k * (t_start - t_grid))  # coordinate-wise product of arrays
        eta = alpha_interp * G_t_T  # coordinate-wise product of arrays
        beta2 = beta_interp * G_t_T  # coordinate-wise product of arrays
        delta = a * eta  # coordinate-wise product of arrays

        term0 = alpha_interp * beta2 * (q ** 2)
        term1 = kappa1 - kappa2 * q + 2.0 * (kappa2 + alpha_interp * beta2) * q
        term2 = kappa2 + alpha_interp * beta2

        return a, delta, term0, term1, term2, beta_interp, volvol_interp


@dataclass
class MultiFactRateLogSvParams(ModelParams):
    """Implementation of model params class for multi-factor model"""
    sigma0: float
    theta: float
    kappa1: float
    kappa2: float
    beta: TermStructure  # beta is term-structure for each driver X_i
    volvol: TermStructure  # residual vol-vol is same for all factors
    # volatilities are term-structure as A = A_t
    A: np.ndarray  # (d,) -- vector matrix of volatilities of key-term forward rates
    # correlations are flat in time
    R: np.ndarray  # (d*d) -- posdef matrix of correlations of key-term forward rates
    basis: Union[NelsonSiegel, CheyettePEND]  # Nelson-Siegel parametrization
    # B: np.ndarray  # (d*d) -- matrix where rows are basis functions
    ccy: str
    vol_interpolation: str = "DIRECT"
    q: float = None

    @classmethod
    def make_A_2d(cls, A: np.ndarray, ts: np.ndarray):
        if A.ndim == 1:
            A_ = np.zeros((ts.size - 1, A.shape[0]))
            for idx, t in enumerate(A_):
                A_[idx, :] = A
        elif A.ndim == 2:
            A_ = A
        else:
            raise NotImplementedError
        return A_

    def __post_init__(self):
        self.key_terms = self.basis.key_terms
        # term-structure times for beta and vol-vol must be consistent
        assert np.all(self.beta.ts == self.volvol.ts)
        # if A is 2D => make is 3D
        self.A = MultiFactRateLogSvParams.make_A_2d(self.A, self.beta.ts)
        assert self.A.shape[0] == self.beta.ts.size - 1  # exclude t=0
        # number of terms must be same as number of factors, equal to 3
        assert len(self.key_terms) == self.basis.nb_factors
        # number of factors in multi-factor term-structure = number of factors in Nelson-Siegel
        assert self.beta.xs.shape[1] == self.basis.nb_factors
        assert self.A.shape[1] == self.basis.nb_factors

        if self.vol_interpolation not in ["BY_YIELD", "DIRECT"]:
            raise NotImplementedError(f"Wrong vol interpolation type")

        # TODO: check if definition of cholesky decomposion is same assume,
        # TODO: in numpy resuling R is lower triangular

        C = np.zeros((self.A.shape[0], self.A.shape[1], self.A.shape[1]))
        M = np.zeros((self.A.shape[0], self.A.shape[1], self.A.shape[1]))
        Omega = np.zeros((self.beta.ts.size - 1, self.basis.nb_aux_factors))
        # interpolation of the volatility matrix is tedious
        for idx, Ai in enumerate(self.A):
            # if Ai.ndim == 1:
            #     Ai = np.reshape(Ai, (Ai.shape[0], 1))
            if self.vol_interpolation == "BY_YIELD":
                Ci = self.calc_factor_vols(Ai)
            else:
                raise NotImplementedError
            Mi = np.dot(Ci, np.transpose(Ci))
            Omegai = self.basis.calc_Omega(Mi)
            C[idx, :] = Ci
            M[idx, :] = Mi
            Omega[idx, :] = Omegai
        self.C = C
        self.M = M
        self.Omega = Omega
        self.ts = self.beta.ts

    #############################################################################
    #                     Analytics under annuity measure Q^A for swaptions
    #############################################################################
    def calc_QA_mean_states(self,
                            expiry: float,
                            tenor: float,
                            t_grid: np.ndarray,
                            x0: np.ndarray,
                            y0: np.ndarray
                            ) -> (np.ndarray, np.ndarray):
        """Calculate Q^A mean \\bar X_t and \\bar Y_t for multi-factor model at times t_grid"""
        ts_sw = get_default_swap_term_structure(expiry=expiry, tenor=tenor)

        def func_mean_rhs(t: float,
                          arg: np.ndarray,
                          ts_sw_: np.ndarray,
                          kappa1: float,
                          kappa2: float,
                          theta: float) -> np.ndarray:
            # evolve vector \\bar X, vector \\barY and scalar vol
            sz_X = self.basis.nb_factors
            sz_Y = self.basis.nb_aux_factors
            sz = sz_X + sz_Y + 1
            res = np.zeros((sz,))
            x, y, sigma = arg[:sz_X], arg[sz_X:sz_X + sz_Y], arg[-1]
            D_X = self.basis.get_generating_matrix()
            D_Y = self.basis.get_aux_generating_matrix()
            idx_t = bracket(self.ts[1:], t, False)
            M_t = self.M[idx_t]
            Omega_t = self.Omega[idx_t]
            beta_t = self.beta.pw_const(t)
            C_t = self.C[idx_t]
            loga_der = 1.0 / self.basis.annuity(t, ts_sw_, x, y, self.ccy, 0) * self.basis.annuity(t, ts_sw_, x, y,
                                                                                                   self.ccy, 1)[0,:]
            # M_t must be matrix, L_X must be a vector column
            res[:sz_X] = np.dot(D_X, x) + sigma ** 2 * np.dot(M_t, loga_der)
            res[sz_X:sz_X+sz_Y] = np.dot(D_Y, y) + sigma ** 2 * Omega_t
            # (beta_t)^T * C^T * L_X is a quadratic form evaluation
            vol_adj = np.dot(np.dot(beta_t, np.transpose(C_t)), loga_der)
            res[-1] = (kappa1 + kappa2 * sigma) * (theta - sigma) + sigma ** 2 * vol_adj

            return res

        sz_X = self.basis.nb_factors
        sz_Y = self.basis.nb_aux_factors
        ode_init = np.concatenate((x0, y0, np.array([self.sigma0])))
        # solve ODE with initial conditions
        oderesult = solve_ivp(fun=func_mean_rhs, t_span=(0, expiry), t_eval=t_grid, y0=ode_init,
                              args=(ts_sw, self.kappa1, self.kappa2, self.theta),
                              dense_output=False)
        mx = np.transpose(oderesult.y[:sz_X, :])
        my = np.transpose(oderesult.y[sz_X:sz_X+sz_Y, :])
        return mx, my


    def transform_QA_params(self,
                            expiry: float,
                            tenor: float,
                            t_grid: np.ndarray,
                            x0: np.ndarray = None,
                            y0: np.ndarray = None
                            ) -> Tuple[np.ndarray, ...]:
        """Calculate a, kappa_0, kappa_1, kappa_2, beta, volvol under Q^A for multi-factor model at times t_grid"""
        if x0 is None:
            x0 = np.zeros((self.basis.get_nb_factors()))
        else:
            assert x0.shape == (self.basis.get_nb_factors(),)
        if y0 is None:
            y0 = np.zeros((self.basis.get_nb_aux_factors()))
        else:
            assert y0.shape == (self.basis.get_nb_aux_factors(),)

        theta = self.theta
        kappa1 = self.kappa1
        kappa2 = self.kappa2
        self.q = self.theta
        q = self.q

        # interpolate a, beta, vol-vol from coarse grid to finer one
        ts_sw = get_default_swap_term_structure(expiry=expiry, tenor=tenor)
        if expiry not in t_grid:
            raise ValueError(f"expiry must be in grid")
        idx_ttm = np.where(t_grid == expiry)[0][0]
        t_grid = t_grid[:idx_ttm + 1]

        mx_grid, my_grid = self.calc_QA_mean_states(expiry=expiry, tenor=tenor, t_grid=t_grid, x0=x0, y0=y0)
        # mx_grid, my_grid = np.zeros((t_grid.size, self.basis.get_nb_factors())), np.zeros((t_grid.size, self.basis.get_nb_aux_factors()))
        swap_grad = np.ones((t_grid.size, self.basis.nb_factors)) * np.nan
        ann = np.ones_like(t_grid) * np.nan
        loga_der = np.ones((t_grid.size, self.basis.nb_factors)) * np.nan
        for idx, (t, mx, my) in enumerate(zip(t_grid, mx_grid, my_grid)):
            swap_grad[idx, :] = self.basis.swap_rate(t, ts_sw, mx, my, ccy=self.ccy)[1]
            ann[idx] = self.basis.annuity(t, ts_sw, mx, my, m=0, ccy=self.ccy)
            ann_der1 = self.basis.annuity(t, ts_sw, mx, my, m=1, ccy=self.ccy)
            loga_der[idx, :] = ann_der1 / ann[idx]

        # interpolate a, beta, vol-vol from coarse grid to finer one
        a_interp = np.ones((t_grid.size, self.basis.nb_factors)) * np.nan
        beta_interp = np.ones((t_grid.size, self.basis.nb_factors)) * np.nan
        volvol_interp = np.ones_like(t_grid) * np.nan
        beta2_interp = np.ones_like(t_grid) * np.nan
        for idx, t in enumerate(t_grid):
            idx_t = bracket(self.ts[1:], t, throw_if_not_found=True)
            beta_interp[idx, :] = self.beta.xs[idx_t]
            volvol_interp[idx] = self.volvol.xs[idx_t]
            a_interp[idx, :] = np.dot(swap_grad[idx, :], self.C[idx_t])
            beta2_interp[idx] = np.dot(beta_interp[idx, :], np.dot(np.transpose(self.C[idx_t]), loga_der[idx, :]))

        term0 = beta2_interp * theta * theta
        term1 = kappa1 - kappa2 * theta + 2.0 * (kappa2-beta2_interp)*theta
        term2 = kappa2 - beta2_interp

        return a_interp, term0, term1, term2, beta_interp, volvol_interp, ts_sw

    #############################################################################
    #                     Analytics under terminal measure Q^T for mid-curves
    #############################################################################
    def transform_QT_params(self,
                            expiry: float,
                            t_start: float,
                            t_end: float,
                            t_grid: np.ndarray) -> Tuple[np.ndarray, ...]:
        # t_start, t_end = get_futures_fixing_and_pmt(expiry=expiry, lag=lag)
        assert expiry <= t_start < t_end
        theta = self.theta
        kappa1 = self.kappa1
        kappa2 = self.kappa2
        # self.q = self.theta
        q = self.theta if self.q is None else self.q

        if expiry not in t_grid:
            raise ValueError(f"expiry must be in grid")
        idx_ttm = np.where(t_grid == expiry)[0][0]
        t_grid = t_grid[:idx_ttm + 1]

        a_interp = np.ones((t_grid.size, self.basis.nb_factors)) * np.nan
        beta_interp = np.ones((t_grid.size, self.basis.nb_factors)) * np.nan
        eta_interp = np.ones((t_grid.size, self.basis.nb_factors)) * np.nan
        volvol_interp = np.ones_like(t_grid) * np.nan
        delta_interp = np.ones_like(t_grid) * np.nan

        term0 = np.ones_like(t_grid) * np.nan
        term1 = np.ones_like(t_grid) * np.nan
        term2 = np.ones_like(t_grid) * np.nan

        # interpolate a, beta, vol-vol from coarse grid to finer one
        for idx, t in enumerate(t_grid):
            idx_t = bracket(self.ts[1:], t, throw_if_not_found=True)
            beta_interp[idx, :] = self.beta.xs[idx_t]
            volvol_interp[idx] = self.volvol.xs[idx_t]
            tau_end = t_end - t
            tau_start = t_start - t
            tau_exp = expiry - t
            B_P_end = self.basis.bond_coeffs(tau_end)[0]
            B_P_start = self.basis.bond_coeffs(tau_start)[0]
            B_P_exp = self.basis.bond_coeffs(tau_exp)[0]
            a_interp[idx, :] = np.dot(np.transpose(self.C[idx_t]), B_P_end-B_P_start)
            eta_interp[idx, :] = np.dot(np.transpose(self.C[idx_t]), B_P_exp)

            beta_x_eta = np.dot(beta_interp[idx, :], eta_interp[idx, :])

            term0[idx] = -beta_x_eta * q**2
            term1[idx] = kappa1 - kappa2 * q + 2.0*(kappa2 + beta_x_eta)
            term2[idx] = kappa2 + beta_x_eta

        return a_interp, eta_interp, term0, term1, term2, beta_interp, volvol_interp

    def check_QT_kappa2(self, t_start: float, t_end: float = None) -> bool:
        if t_end is None:
            t_end = t_start + 0.25
        expiry = t_start
        t_grid = generate_ttms_grid(np.array([expiry]))
        a_interp, delta_interp, term0, term1, term2, beta_interp, volvol_interp = self.transform_QT_params(expiry=expiry, t_start=t_start, t_end=t_end, t_grid=t_grid)
        return np.all(term2 > 0.0)

    def check_QA_kappa2(self, expiry: float, tenor: float) -> bool:
        t_grid = generate_ttms_grid(np.array([expiry]))

        a_interp, term0, term1, term2, beta_interp, volvol_interp, _ = self.transform_QA_params(expiry=expiry,
                                                                                                tenor=tenor,
                                                                                                t_grid=t_grid)
        return np.all(term2 > 0.0)


    def reduce(self, ids: List[str]):
        ttms = [MultiFactRateLogSvParams.get_frac(id) for id in ids]
        assert set(ttms) <= set(self.ts)
        indices = np.in1d(self.ts, ttms).nonzero()[0] - 1
        ts_indices = np.concatenate(([0], indices+1))
        assert np.all(indices >= 0)

        param = MultiFactRateLogSvParams(sigma0=self.sigma0,
                                         theta=self.theta,
                                         kappa1=self.kappa1,
                                         kappa2=self.kappa2,
                                         beta=TermStructure(self.beta.ts[ts_indices], self.beta.xs[indices]),
                                         volvol=TermStructure(self.volvol.ts[ts_indices], self.volvol.xs[indices]),
                                         A=self.A[indices, :],
                                         R=self.R,
                                         basis=self.basis,
                                         ccy=self.ccy,
                                         vol_interpolation=self.vol_interpolation,
                                         q=self.q)
        return param

    def update_params(self, idx: int, A_idx: np.ndarray = None, beta_idx: np.ndarray = None, volvol_idx: float = None,
                      kappa1: float = None, kappa2: float = None,
                      sigma0: float = None):
        nb_factors = self.basis.get_nb_factors()
        if A_idx is not None:
            assert A_idx.shape == (nb_factors,)
            self.A[idx, :] = A_idx
        if beta_idx is not None:
            assert beta_idx.shape == (nb_factors,)
            self.beta.xs[idx, :] = beta_idx
        if volvol_idx is not None:
            self.volvol.xs[idx] = volvol_idx
        if kappa1 is not None:
            self.kappa1 = kappa1
        if kappa2 is not None:
            self.kappa2 = kappa2
        if sigma0 is not None:
            self.sigma0 = sigma0


        self.__post_init__()

    @classmethod
    def get_frac(cls, id: str) -> float:
        if id == '3m':
            return 0.25
        elif id == '6m':
            return 0.5
        elif id == '1y':
            return 1.0
        elif id == '2y':
            return 2.0
        elif id == '3y':
            return 3.0
        elif id == '4y':
            return 4.0
        elif id == '5y':
            return 5.0
        elif id == '7y':
            return 7.0
        elif id == '10y':
            return 10.0
        elif id == '31d':
            return 31.0/365.0
        elif id == '40d':
            return 40.0/365.0
        elif id == '66d':
            return 66.0/365
        elif id == '75d':
            return 75.0/365
        elif id == '84d':
            return 84.0/365
        elif id == '87d':
            return 87.0/365
        elif id == '103d':
            return 103.0/365
        elif id == '156d':
            return 156.0/365
        elif id == '194d':
            return 194.0/365
        else:
            raise NotImplementedError(f"id not found")


    def calc_factor_vols(self, yield_vols: np.ndarray) -> np.ndarray:
        assert yield_vols.ndim == 1 and yield_vols.shape[0] == self.basis.get_nb_factors()
        B = self.basis.get_matrix_B()
        R_chol = np.linalg.cholesky(self.R)
        factor_vol = np.dot(np.dot(np.linalg.inv(B), np.diag(yield_vols)), R_chol)
        # factor_vol = np.dot(np.diag(yield_vols), R_chol)
        return factor_vol

    def calc_factor_vols_dln(self, yield_vols: np.ndarray,
                             yields: np.ndarray,
                             b_dln: np.ndarray,
                             nb_path: int):

        nb_factors = self.basis.get_nb_factors()
        assert yield_vols.ndim == 1 and yield_vols.shape[0] == nb_factors
        assert b_dln.shape == yield_vols.shape
        assert yields.shape == (nb_path, nb_factors)
        B = self.basis.get_matrix_B()
        R_chol = np.linalg.cholesky(self.R)
        inv_B = np.linalg.inv(B)
        A_DLN = np.zeros((nb_path, nb_factors, nb_factors))
        factor_vol = np.zeros((nb_path, nb_factors, nb_factors))
        b_times_y = np.multiply(yields, b_dln)
        for i, b_times_y_ith in enumerate(b_times_y):
            A_DLN[i, :, :] = np.diag(yield_vols + b_times_y_ith)
            factor_vol[i, :, :] = np.dot(np.dot(inv_B, A_DLN[i, :, :]), R_chol)

        return factor_vol

